/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author Joan Segura <joan.segura@rcsb.org>
 * @author Yana Rose <yana.rose@rcsb.org>
 * @author Sebastian Bittrich <sebastian.bittrich@rcsb.org>
 */
/**
 * Copyright (C) 2024, Protein Bioinformatics Research Group, HUN-REN RCNS
 *
 * Licensed under CC BY-NC 4.0, see LICENSE file for more info.
 *
 * @author Gabor Tusnady <tusnady.gabor@ttk.hu>
 * @author Csongor Gerdan <gerdan.csongor@ttk.hu>
 */

import { Viewer, ViewerProps } from '@rcsb/rcsb-molstar/build/src/viewer';
import { TmDetColorThemeProvider, updateSiteColors } from '../../TmFv3DApp/tmdet-extension/tmdet-color-theme';
import { MembraneOrientationPreset } from '../../TmFv3DApp/tmdet-extension/behavior';
import { TmDetLabelProvider } from '../../TmFv3DApp/tmdet-extension/labeling';
import { BuiltInTrajectoryFormat } from 'molstar/lib/mol-plugin-state/formats/trajectory';
import { PresetProps } from '@rcsb/rcsb-molstar/build/src/viewer/helpers/preset';
import { Mat4 } from 'molstar/lib/commonjs/mol-math/linear-algebra';
import { TrajectoryHierarchyPresetProvider } from 'molstar/lib/mol-plugin-state/builder/structure/hierarchy-preset';
import { registerRegionDescriptorData } from '../../TmFv3DApp/UniTmpHelper';
import { TmDetRcsbPreset } from '../TmTrajectoryHierarchyPreset';
import { Expression } from 'molstar/lib/mol-script/language/expression';
import { MolScriptBuilder } from 'molstar/lib/mol-script/language/builder';
import { StateTransforms } from 'molstar/lib/mol-plugin-state/transforms';
import { Color } from 'molstar/lib/mol-util/color';

import { StructureElement, Structure, StructureSelection, QueryContext } from 'molstar/lib/mol-model/structure';
import { OrderedSet } from 'molstar/lib/mol-data/int';
import { setStructureOverpaint } from 'molstar/lib/mol-plugin-state/helpers/structure-overpaint';
import { Script } from 'molstar/lib/mol-script/script';
import { compile } from 'molstar/lib/mol-script/runtime/query/base';
import { Overpaint } from 'molstar/lib/mol-theme/overpaint';
import { createStructureRepresentationParams } from 'molstar/lib/mol-plugin-state/helpers/structure-representation-params';
import { ColorNames } from 'molstar/lib/mol-util/color/names';

export class TmViewerStandalone extends Viewer {

    constructor(elementOrId: string | HTMLElement, props: Partial<ViewerProps> = {}, withTmDetBehaviour: boolean = true) {

        super(elementOrId, props);

        if (withTmDetBehaviour) {
            this.initBehaviour();
        }
    }

    public async initBehaviour() {
        let tree = this.plugin.state.behaviors.build();


        this.plugin.builders.structure.representation.registerPreset(MembraneOrientationPreset);
        this.plugin.representation.structure.themes.colorThemeRegistry.add(TmDetColorThemeProvider);
        this.plugin.managers.lociLabels.addProvider(TmDetLabelProvider);

        await this.plugin.runTask(this.plugin.state.behaviors.updateTree(tree, { doNotUpdateCurrent: true, doNotLogTiming: true }));
    }

    loadStructureFromUrl<P, S>(url: string, format: BuiltInTrajectoryFormat, isBinary: boolean, config?: {
        props?: PresetProps & {
            dataLabel?: string;
        };
        matrix?: Mat4;
        reprProvider?: TrajectoryHierarchyPresetProvider<P, S>;
        params?: P;
    }): Promise<any> {
        return super.loadStructureFromUrl(url, format, isBinary, config);
    }

    async lociGetter(structure: Structure) {
        // const structure = this.plugin.managers.structure.hierarchy.current.models[0].structures[0].cell;
        // const unit = structure.obj!.data!.unitMap.get(0)
        // structure.obj!.data.entityIndices
        // StructureElement.Loci(structure.obj!.data, [{ unit: unit, indices: Array.from(unit.elements) }]);
        // console.log(Structure.Loci());
        let indices = OrderedSet.ofRange(10 as StructureElement.UnitIndex, 100  as StructureElement.UnitIndex);
        // let indices = OrderedSet.ofSortedArray([ 15 as StructureElement.UnitIndex ]);
        return StructureElement.Loci(structure, [
            { unit: structure.unitMap.get(0), indices: indices },
            { unit: structure.unitMap.get(1), indices: indices },
            { unit: structure.unitMap.get(2), indices: indices }
        ]);
    }

    chainSelection(auth_asym_id: string) {
        return MolScriptBuilder.struct.generator.atomGroups({
            'chain-test': MolScriptBuilder.core.rel.eq([MolScriptBuilder.struct.atomProperty.macromolecular.auth_asym_id(), auth_asym_id])
        });
    }

    async setChainColorOLD(labelAsymId: string) {

        const state = this.plugin.state.data;

        const selected = state.selectQ(q => q.ofTransformer(StateTransforms.Model.StructureFromModel));
        //const str = this.plugin.managers.structure.hierarchy.current.structures[0];
        if (selected.length == 0) {
            console.log('Structure not found');
            return;
        }
        let structure = selected[0];
        console.log('STRUCT1', structure);
        structure = this.plugin.managers.structure.hierarchy.current.models[0].structures[0].cell;
        console.log('STRUCT2', structure);

        // const update = this.plugin.build().to(structure);  //state.build().to(structure);
        const query: Expression =
            MolScriptBuilder.struct.generator.atomGroups({
                'chain-test': MolScriptBuilder.core.rel.eq([labelAsymId, MolScriptBuilder.ammp('label_asym_id')])
        });

        const s = await this.plugin.build()
            .to(this.plugin.managers.structure.hierarchy.current.structures[0].cell)
            .apply(StateTransforms.Model.StructureSelectionFromExpression, { expression: query }).commit();
        // const rangeSel = Script.getStructureSelection(Q => Q.struct.modifier.union([
        //     Q.struct.generator.atomGroups({
        //         // I am not sure this test will be enough, or uses the correct property, but something along these lines should do it
        //         'chain-test': Q.core.rel.eq([Q.struct.atomProperty.macromolecular.label_asym_id(), 'A']),
        //     })
        // ]), );
        const compiledQuery = compile<StructureSelection>(query);
        const xs = this.plugin.managers.structure.hierarchy.current.structures;
        const selections = xs.map(s => StructureSelection.toLociWithCurrentUnits(compiledQuery(new QueryContext(s.cell.obj!.data))));

        console.log(selections);


        const component = await this.plugin.builders.structure.tryCreateComponentFromExpression(structure, query, 'hello', { label: 'CHAIN-A' });
        console.log(component);
        // const update = this.plugin.build().to(component!.cell!);
        // update //.apply(StateTransforms.Model.StructureSelectionFromExpression, { label: 'CHAIN-A', expression: query })
        //     .apply(StateTransforms.Representation.StructureRepresentation3D, createStructureRepresentationParams(this.plugin, update.selector.data,
        //         { type: 'cartoon', typeParams: { alpha: 1 },
        //         color: 'uniform', colorParams: { value: Color.fromArray([255,0,0], 0) }  }))
        // const result = await update.commit();
        // console.log('Result:', result);

        console.log('Component Group:', this.plugin.managers.structure.hierarchy.currentComponentGroups);
        // await this.plugin.builders.structure.representation.addRepresentation(structure, {
        //     type: 'cartoon',
        //     color: 'uniform',
        //     colorParams: {value: 0xd3d3d3}, // gray80
        //     typeParams: {detail: 1}
        // });

        // TODO: const structureData = component!.cell!.obj!.data;

        //setStructureOverpaint(this.plugin, [ structure!.obj!.data ], Color.fromRgb(255, 0, 0), this.lociGetter, undefined);
        // console.log(component);
        console.log('component cell', component!.cell!);
        const structRef = this.plugin.managers.structure.hierarchy.current.refs;
        let comps: any[] = [];

        // const bundle = StructureElement.Bundle.fromLoci(selections[0]);
        // await this.plugin.build().to(component!.cell!)
        //     .apply(StateTransforms.Representation.OverpaintStructureRepresentation3DFromScript,
        //         {
        //             layers: [{
        //                 script: query,
        //                 color: Color.fromRgb(255, 0, 0),
        //                 clear: false
        //             }]
        //         })
        //     .commit();


        this.plugin.managers.structure.component.currentStructures.forEach(val => comps = comps.concat(...val.components));
setStructureOverpaint(this.plugin, comps, Color.fromRgb(255, 0, 0), this.lociGetter, undefined);
        console.log('Components', comps);
        // console.log('component ref', this.plugin.managers.structure.component.;
    }

    async setChainColor(labelAsymId: string) {

        const state = this.plugin.state.data;

        const selected = state.selectQ(q => q.ofTransformer(StateTransforms.Model.StructureFromModel));
        //const str = this.plugin.managers.structure.hierarchy.current.structures[0];
        if (selected.length == 0) {
            console.log('Structure not found');
            return;
        }

        const query: Expression =
            MolScriptBuilder.struct.generator.atomGroups({
                'chain-test': MolScriptBuilder.core.rel.eq([labelAsymId, MolScriptBuilder.ammp('label_asym_id')])
        });

        // const update = this.plugin.build().to(structure)
        const update = this.plugin.build().to(Array.from(state.cells.values())[13]);
        update //.apply(StateTransforms.Model.StructureSelectionFromExpression, { label: 'CHAIN-A', expression: query })
            // .apply(StateTransforms.Representation.StructureRepresentation3D, createStructureRepresentationParams(this.plugin, update.selector.data,
            //     { type: 'cartoon', typeParams: { alpha: 1 },
            //     color: 'uniform', colorParams: { value: Color.fromArray([255,0,0], 0) }  }))
            .apply(StateTransforms.Representation.OverpaintStructureRepresentation3DFromScript, {
                layers: [
                    {
                        script:  Script(
                            `(sel.atom.res (in-range atom.resno 10 100))`,
                            'mol-script'
                        ),
                        color: ColorNames.red,
                        clear: false

                    }
                ]
            });
        const result = await update.commit();
        console.log('Result:', result);

        // await this.plugin.build().to(component!.cell!)
        //     .apply(StateTransforms.Representation.OverpaintStructureRepresentation3DFromScript,
        //         {
        //             layers: [{
        //                 script: query,
        //                 color: Color.fromRgb(255, 0, 0),
        //                 clear: false
        //             }]
        //         })
        //     .commit();
    }

    async loadWithUNITMPMembraneRepresentation(params: any) {
        setTimeout(() => { this.plugin.clear(); }, 100); // clear scene after some delay
        updateSiteColors(params.side1);

        setTimeout(() => { (async () => {
            await registerRegionDescriptorData(params.regionDescriptorUrl, params.side1);

            // load structure
            this.loadStructureFromUrl(params.structureUrl, 'mmcif', false, { reprProvider: TmDetRcsbPreset });

        })(); }, 500);
    }

}
